/*
 * web_queue.c - An empty file. Use it if you wish, don't if you don't.
 * 
 */

/*<solution>*/

#include <config.h>

#include <stdlib.h>
#include <assert.h>
#include <sthread.h>

#include <web_queue.h>

struct _web_queue_elem {
  int conn;
  struct _web_queue_elem *next;
};
typedef struct _web_queue_elem* web_queue_elem_t;

static web_queue_elem_t free_list;
static sthread_mutex_t free_list_lock;

struct _web_queue {
  web_queue_elem_t head;
  web_queue_elem_t tail;
  int size;
};

/* Create a new, empty queue. Asserts against error. */
web_queue_t web_new_queue() {
  web_queue_t queue;

  queue = (web_queue_t)malloc(sizeof(struct _web_queue));
  assert(queue != NULL);
  
  queue->head = queue->tail = NULL;
  queue->size = 0;

  free_list_lock = sthread_mutex_init();
  
  return queue;
}

/* Destroy the given queue. Asserts that the queue is empty. */
void web_free_queue(web_queue_t queue) {
  assert(queue->size == 0);
  free(queue);
}


/* Add the given thread to the end of the queue */
void web_enqueue(web_queue_t queue, int conn) {
  web_queue_elem_t elem;
  sthread_mutex_lock(free_list_lock);
  elem = free_list;
  if (elem == NULL)
    elem = (web_queue_elem_t)malloc(sizeof(struct _web_queue_elem));
   else
    free_list = free_list->next;
  sthread_mutex_unlock(free_list_lock);
  
  assert(elem != NULL);
  
  elem->next = NULL;
  elem->conn = conn;

  if (queue->tail != NULL) {
    queue->tail->next = elem;
  } else {
    assert(queue->head == NULL);
    queue->head = elem;
  }
  queue->tail = elem;

  queue->size++;
}

/* Return, and remove, the next thread from the queue, or NULL
 * if queue is empty */
int web_dequeue(web_queue_t queue) {
  web_queue_elem_t head;
  int conn;
  
  if (queue->head == NULL)
    return 0;

  head = queue->head;
  
  conn = head->conn;
  if (head->next == NULL) {
    assert(queue->size == 1);
    assert(head == queue->tail);
    queue->tail = NULL;
  }
  queue->head = head->next;

  /* Return to free list */
  sthread_mutex_lock(free_list_lock);
  head->next = free_list;
  free_list = head;
  sthread_mutex_unlock(free_list_lock);
  
  head = NULL;
  
  queue->size--;

  return conn;
}

/* Return the number of threads currently in the queue */
int web_queue_size(web_queue_t queue) {
  return queue->size;
}

/* Return true if queue has no threads, false otherwise */
int web_queue_is_empty(web_queue_t queue) {
  return (queue->size == 0);
}

/*</solution>*/
